@@ -5,14 +5,16 @@ Contains the functions needed for the weight computation of MCF
 -*- coding: utf-8 -*-
 """
 from concurrent import futures
+
 import numpy as np
 import pandas as pd
 from scipy import sparse
 import ray
+
 from mcf import general_purpose as gp
 from mcf import general_purpose_system_files as gp_sys
-from mcf import general_purpose_mcf as gp_mcf
-from mcf import mcf_forest_functions as mcf_forest
+from mcf import mcf_general_purpose as mcf_gp
+from mcf import mcf_forest_add_functions as mcf_forest_add
 
 
 def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
@@ -37,7 +39,6 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
     cl_dat : N_y x 1 Numpy array. Cluster number.
     w_dat: N_y x 1 Numpy array. Sampling weights (if used).
     """
-    # gp_sys.clean_futures()
     if c_dict['with_output'] and c_dict['verbose']:
         print('\nObtaining weights from estimated forest')
     data_x = pd.read_csv(x_file)
@@ -45,34 +46,24 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
     del data_x
     data_y = pd.read_csv(y_file)
     y_dat = data_y[v_dict['y_name']].to_numpy()
-    if regrf:
-        d_dat = None
-    else:
-        d_dat = data_y[v_dict['d_name']].to_numpy()
+    d_dat = None if regrf else data_y[v_dict['d_name']].to_numpy()
     if not regrf:
         d_dat = np.int16(np.round(d_dat))
-    n_x = len(x_dat)
-    n_y = len(y_dat)
-    if c_dict['cluster_std']:
-        cl_dat = data_y[v_dict['cluster_name']].to_numpy()
-    else:
-        cl_dat = 0
-    if c_dict['w_yes']:
-        w_dat = data_y[v_dict['w_name']].to_numpy()
-    else:
-        w_dat = 0
+    n_x, n_y = len(x_dat), len(y_dat)
+    cl_dat = (data_y[v_dict['cluster_name']].to_numpy()
+              if c_dict['cluster_std'] else 0)
+    w_dat = data_y[v_dict['w_name']].to_numpy() if c_dict['w_yes'] else 0
     x_bala = 0
     if not regrf:
         if c_dict['balancing_test_w']:
             x_bala = data_y[v_dict['x_balance_name']].to_numpy()
     del data_y
-    empty_leaf_counter = 0
-    merge_leaf_counter = 0
+    empty_leaf_counter = merge_leaf_counter = 0
     if c_dict['no_parallel'] < 1.5:
         maxworkers = 1
     else:
         if c_dict['mp_automatic']:
-            maxworkers = gp_mcf.find_no_of_workers(c_dict['no_parallel'],
+            maxworkers = mcf_gp.find_no_of_workers(c_dict['no_parallel'],
                                                    c_dict['sys_share'])
         else:
             maxworkers = c_dict['no_parallel']
@@ -81,10 +72,11 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
     if maxworkers == 1 or c_dict['mp_with_ray']:
         mp_over_boots = False
     else:
-        if regrf:
-            mp_over_boots = False
-        else:
-            mp_over_boots = bool(c_dict['mp_type_weights'] == 2)
+        mp_over_boots = (False if regrf
+                         else bool(c_dict['mp_type_weights'] == 2))
+    no_of_treat = (len(c_dict['ct_grid_w_val'])
+                   if c_dict['d_type'] == 'continuous'
+                   else c_dict['no_of_treat'])
     if maxworkers == 1 or mp_over_boots:
         weights = initialise_weights(c_dict, n_x, n_y, regrf)
         split_forest = False
@@ -95,7 +87,7 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
             if c_dict['with_output'] and c_dict['verbose']:
                 gp.share_completed(idx+1, n_x)
             if c_dict['weight_as_sparse']:
-                for d_idx in range(c_dict['no_of_treat']):
+                for d_idx in range(no_of_treat):
                     indices = results_fut_idx[1][d_idx][0]
                     weights_obs = results_fut_idx[1][d_idx][1]
                     weights[d_idx][idx, indices] = weights_obs
@@ -106,22 +98,20 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
             empty_leaf_counter += results_fut_idx[2]
             merge_leaf_counter += results_fut_idx[3]
         if c_dict['weight_as_sparse']:
-            weights = weights_to_csr(weights, c_dict['no_of_treat'], regrf)
+            weights = weights_to_csr(weights, no_of_treat, regrf)
     else:
-        no_of_splits_i = maxworkers
-        max_size_i = 1000
+        no_of_splits_i, max_size_i = maxworkers, 1000
         if n_x / no_of_splits_i > max_size_i:
             while True:
                 no_of_splits_i += maxworkers
                 if n_x / no_of_splits_i <= max_size_i:
                     break
         if c_dict['with_output'] and c_dict['verbose']:
-            print()
-            print('Operational characteristics of weight estimation I:')
-            print('Number of workers {:2}'.format(maxworkers))
-            print('Number of observation chunks: {:5}'.format(no_of_splits_i))
-            print('Average # of observations per chunck: {:5.2f}'.format(
-                n_x / no_of_splits_i))
+            print('\nOperational characteristics of weight estimation I:')
+            print(f'Number of workers {maxworkers:2}')
+            print(f'Number of observation chunks: {no_of_splits_i:5}')
+            print('Average # of observations per chunck:',
+                  f' {n_x / no_of_splits_i:5.2f}')
         all_idx_split = np.array_split(range(n_x), no_of_splits_i)
         split_forest = False
         if not c_dict['mp_with_ray']:
@@ -129,11 +119,10 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                 no_of_boot_splits = c_dict['mp_weights_tree_batch']
                 split_forest = True
                 if c_dict['with_output'] and c_dict['verbose']:
-                    print()
-                    print('User determined number of tree batches')
+                    print('\nUser determined number of tree batches')
             elif c_dict['mp_weights_tree_batch'] == 0:  # Automatic # of batch
                 size_of_forest_mb = gp_sys.total_size(forest) / (1024 * 1024)
-                no_of_boot_splits = gp_mcf.no_of_boot_splits_fct(
+                no_of_boot_splits = mcf_gp.no_of_boot_splits_fct(
                     size_of_forest_mb, maxworkers, c_dict['with_output'])
                 if no_of_boot_splits > 1:
                     split_forest = True
@@ -145,10 +134,9 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                                             no_of_boot_splits)
             total_bootstraps = c_dict['boot']
             if c_dict['with_output'] and c_dict['verbose']:
-                print('Number of bootstrap chunks: {:5}'.format(
-                    no_of_boot_splits))
-                print('Average # of bootstraps per chunck: {:5.2f}'.format(
-                    c_dict['boot']/no_of_boot_splits))
+                print(f'Number of bootstrap chunks: {no_of_boot_splits:5}')
+                print('Average # of bootstraps per chunck:',
+                      f' {c_dict["boot"]/no_of_boot_splits:5.2f}')
         else:
             boot_indx_list = range(1)
         if (not c_dict['weight_as_sparse']) and split_forest:
@@ -166,14 +154,14 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                 # forest_temp = [forest[boot] for boot in boots_ind]
                 forest_temp = forest[boots_ind[0]:boots_ind[-1]+1]
                 if c_dict['with_output'] and c_dict['verbose'] and b_i == 0:
-                    print('Size of each submitted forest {:6.2f} MB'
-                          .format(gp_sys.total_size(forest_temp)/(1024*1024)))
+                    print('Size of each submitted forest ',
+                          f'{gp_sys.total_size(forest_temp)/(1024*1024):6.2f}',
+                          ' MB')
                 c_dict['boot'] = len(boots_ind)
                 # weights über trees addieren
                 if c_dict['with_output'] and c_dict['verbose']:
                     print()
-                    print('Boot Chunk {:2} of {:2}'.format(
-                        b_i+1, no_of_boot_splits))
+                    print(f'Boot Chunk {b_i+1:2} of {no_of_boot_splits:2}')
                     gp_sys.memory_statistics()
             else:
                 if not c_dict['mp_with_ray']:
@@ -193,10 +181,6 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                               " MB")
                 x_dat_ref = ray.put(x_dat)
                 forest_ref = ray.put(forest)
-                # tasks = [ray_weights_many_obs_i.remote(
-                #     idx_list, n_y, forest_ref, x_dat_ref, d_dat, c_dict, regrf,
-                #     split_forest) for idx_list in all_idx_split]
-                # still_running = list(tasks)
                 still_running = [ray_weights_many_obs_i.remote(
                     idx_list, n_y, forest_ref, x_dat_ref, d_dat, c_dict, regrf,
                     split_forest) for idx_list in all_idx_split]
@@ -208,15 +192,15 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                         if c_dict['with_output'] and c_dict['verbose']:
                             if jdx == 0:
                                 print()
-                                print('   Obs chunk {:2} ({:2})'.format(
-                                    jdx+1, no_of_splits_i), end='')
+                                print(f'   Obs chunk {jdx+1:2}',
+                                      f' ({no_of_splits_i:2})', end='')
                             else:
-                                print(' {:2} ({:2})'.format(
-                                    jdx+1, no_of_splits_i), end='')
+                                print(f' {jdx+1:2} ({no_of_splits_i:2})',
+                                      end='')
                             jdx += 1
                         for idx, val_list in enumerate(results_fut_idx[0]):
                             if c_dict['weight_as_sparse']:
-                                for d_idx in range(c_dict['no_of_treat']):
+                                for d_idx in range(no_of_treat):
                                     indices = results_fut_idx[1][idx][d_idx][0]
                                     weights_obs = results_fut_idx[1][idx][d_idx
                                                                           ][1]
@@ -232,8 +216,6 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                         merge_leaf_counter += results_fut_idx[3]
                 if 'refs' in c_dict['_mp_ray_del']:
                     del x_dat_ref, forest_ref
-                # if 'remote' in c_dict['_mp_ray_del']:
-                #     del tasks
                 if 'rest' in c_dict['_mp_ray_del']:
                     del finished_res, finished
                 if c_dict['_mp_ray_shutdown']:
@@ -252,15 +234,15 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                         if c_dict['with_output'] and c_dict['verbose']:
                             if jdx == 0:
                                 print()
-                                print('   Obs chunk {:2} ({:2})'.format(
-                                    jdx+1, no_of_splits_i), end='')
+                                print(f'   Obs chunk {jdx+1:2}',
+                                      f' ({no_of_splits_i:2})', end='')
                             else:
-                                print(' {:2} ({:2})'.format(
-                                    jdx+1, no_of_splits_i), end='')
+                                print(f' {jdx+1:2} ({no_of_splits_i:2})',
+                                      end='')
                         idx_list = results_fut_idx[0]
                         for idx, val_list in enumerate(idx_list):
                             if c_dict['weight_as_sparse']:
-                                for d_idx in range(c_dict['no_of_treat']):
+                                for d_idx in range(no_of_treat):
                                     indices = results_fut_idx[1][idx][d_idx][0]
                                     weights_obs = results_fut_idx[1][idx][d_idx
                                                                           ][1]
@@ -273,34 +255,32 @@ def get_weights_mp(forest, x_file, y_file, v_dict, c_dict, x_name,
                         empty_leaf_counter += results_fut_idx[2]
                         merge_leaf_counter += results_fut_idx[3]
             if c_dict['weight_as_sparse']:
-                weights = weights_to_csr(weights, c_dict['no_of_treat'],
-                                         regrf)
+                weights = weights_to_csr(weights, no_of_treat, regrf)
             if split_forest:
                 if b_i == 0:  # only in case of splitted forests
                     weights_all = weights
                 else:
-                    for d_idx in range(c_dict['no_of_treat']):
+                    for d_idx in range(no_of_treat):
                         weights_all[d_idx] += weights[d_idx]
                         if regrf:
                             break
                 if c_dict['with_output'] and c_dict['verbose']:
                     print()
-                    gp_mcf.print_size_weight_matrix(
-                        weights_all, c_dict['weight_as_sparse'],
-                        c_dict['no_of_treat'])
+                    mcf_gp.print_size_weight_matrix(
+                        weights_all, c_dict['weight_as_sparse'], no_of_treat)
                     gp_sys.memory_statistics()
     if split_forest:
         c_dict['boot'] = total_bootstraps
-        weights = normalize_weights(weights_all, c_dict['no_of_treat'], regrf,
+        weights = normalize_weights(weights_all, no_of_treat, regrf,
                                     c_dict['weight_as_sparse'], n_x)
     weights = tuple(weights)
     if (((empty_leaf_counter > 0) or (merge_leaf_counter > 0))
             and c_dict['with_output']) and c_dict['verbose']:
         print('\n')
-        print('{:5} observations attributed in merged leaves'.format(
-            merge_leaf_counter))
-        print('{:5} observations attributed to leaf w/o observations'.format(
-            empty_leaf_counter))
+        print(f'{merge_leaf_counter:5} observations attributed in merged',
+              ' leaves')
+        print(f'{empty_leaf_counter:5} observations attributed to leaf w/o',
+              ' observations')
     return weights, y_dat, x_bala, cl_dat, w_dat
 
 
@@ -335,14 +315,15 @@ def weights_to_csr(weights, no_of_treat, regrf):
 
 def initialise_weights(c_dict, n_x, n_y, regrf):
     """Initialise the weights matrix."""
+    no_of_treat = (len(c_dict['ct_grid_w_val'])
+                   if c_dict['d_type'] == 'continuous'
+                   else c_dict['no_of_treat'])
     if c_dict['weight_as_sparse']:
         weights = []
-        for _ in range(c_dict['no_of_treat']):
+        for _ in range(no_of_treat):
             weights.append(sparse.lil_matrix((n_x, n_y), dtype=np.float32))
             if regrf:
                 break
-        # weights = [sparse.lil_matrix((n_x, n_y), dtype=np.float32)
-        #            ] * c_dict['no_of_treat']
     else:
         weights = [None] * n_x
     return weights
@@ -381,8 +362,7 @@ def weights_many_obs_i(idx_list, n_y, forest, x_dat, d_dat, c_dict,
 
     """
     weights = [None] * len(idx_list)
-    empty_leaf_counter = 0
-    merge_leaf_counter = 0
+    empty_leaf_counter = merge_leaf_counter = 0
     for idx, val in enumerate(idx_list):
         results_fut_idx = weights_obs_i(
             val, n_y, forest, x_dat, d_dat, c_dict, regrf, mp_over_boots=False,
@@ -420,22 +400,29 @@ def weights_obs_i(idx, n_y, forest, x_dat, d_dat, c_dict, regrf=False,
     merge_leaf_counter : Int.
 
     """
-    empty_leaf_counter = 0
-    merge_leaf_counter = 0
+    empty_leaf_counter = merge_leaf_counter = 0
+    if c_dict['d_type'] == 'continuous':
+        no_of_treat = len(c_dict['ct_grid_w_val'])
+        d_values = c_dict['ct_grid_w_val']
+        continuous, regrf = True, False
+    else:
+        no_of_treat, d_values = c_dict['no_of_treat'], c_dict['d_values']
+        continuous = False
     if regrf:
         weights_i_np = np.zeros((n_y, 2))
         weights_i = [None]
+        continuous = False
     else:
-        weights_i_np = np.zeros((n_y, c_dict['no_of_treat'] + 1))
-        weights_i = [None] * c_dict['no_of_treat']  # 1 list for each treatment
+        weights_i_np = np.zeros((n_y, no_of_treat + 1))
+        weights_i = [None] * no_of_treat  # 1 list for each treatment
     weights_i_np[:, 0] = np.arange(n_y)  # weight for index of outcomes
     x_dat_i = x_dat[idx, :]
     if mp_over_boots:
         with futures.ProcessPoolExecutor(max_workers=maxworkers) as fpp:
             ret_fut = {fpp.submit(
                 weights_obs_i_inside_boot, forest[boot], x_dat_i, regrf, n_y,
-                c_dict['no_of_treat'], c_dict['d_values'], d_dat):
-                boot for boot in range(c_dict['boot'])}
+                no_of_treat, d_values, d_dat, continuous):
+                    boot for boot in range(c_dict['boot'])}
             for fur in futures.as_completed(ret_fut):
                 result = fur.result()
                 del ret_fut[fur]
@@ -447,68 +434,101 @@ def weights_obs_i(idx, n_y, forest, x_dat, d_dat, c_dict, regrf=False,
     else:
         for boot in range(c_dict['boot']):
             weights_ij_np, empty_leaf = weights_obs_i_inside_boot(
-                forest[boot], x_dat_i, regrf, n_y, c_dict['no_of_treat'],
-                c_dict['d_values'], d_dat)
+                forest[boot], x_dat_i, regrf, n_y, no_of_treat, d_values,
+                d_dat, continuous)
             if empty_leaf:
                 empty_leaf_counter += 1
             else:
                 weights_i_np[:, 1:] += weights_ij_np
-    if empty_leaf_counter == c_dict['boot']:
-        obs_without_leaf = 1
-    else:
-        obs_without_leaf = 0
-    if split_forest:
-        normalize = False
-    else:
-        normalize = True
+    obs_without_leaf = 1 if empty_leaf_counter == c_dict['boot'] else 0
+    normalize = not split_forest
     if regrf:
         weights_i = final_trans(weights_i_np, None, regrf, normalize)
     else:
-        weights_i = final_trans(weights_i_np, c_dict['no_of_treat'], regrf,
-                                normalize)
+        weights_i = final_trans(weights_i_np, no_of_treat, regrf, normalize)
     return idx, weights_i, obs_without_leaf, merge_leaf_counter
 
 
 def weights_obs_i_inside_boot(forest_b, x_dat_i, regrf, n_y, no_of_treat,
-                              d_values, d_dat):
+                              d_values, d_dat, continuous):
     """Allow for MP at bootstrap level (intermediate procedure)."""
-    leaf_id = mcf_forest.get_terminal_leaf_no(forest_b, x_dat_i)
-    if forest_b[leaf_id][14] is None:  # Leave will ignored
-        # if int(leaf_id) % 2 == 0:  # even leaf id
-        #     leaf_id -= 1
-        # else:                      # odd leaf id
-        #     leaf_id += 1
-        empty_leaf = True
-        weights_ij_np = 0
+    # here is the information for the different treatments needed
+    # Problem ist dass man hier für die x alle potential outcomes berechnet
+    # Bei continuous treatments gehören aber verschiedene leafs zu den gleichen
+    # X. Ausserdem werden d=0 und die anderen Werte von d verschieden behandelt
+    # x_dat_i ist eine design matrix, die dann auch die verschiedenen werte von
+    # d enthalten sollte
+    # d = 0 zu setzen scheint aber überflüssig bei der designmatrix bzgl. d
+    if continuous:
+        leaf_id_list = []
+        # We get a list of leafs that contains relevant elements, some of them
+        # may be identical, but this does not matter
+        for treat in d_values[1:]:
+            x_dat_i_t = np.append(x_dat_i, treat)
+            leaf_id = (mcf_forest_add.get_terminal_leaf_no(forest_b,
+                                                           x_dat_i_t))
+            leaf_id_list.append(leaf_id)
+            if forest_b[leaf_id][14] is None:  # Leave will be ignored
+                empty_leaf, weights_ij_np = True, 0
+                # If any of the subleaves is None, then stop ...
+                break
+            empty_leaf = False
+
     else:
-        empty_leaf = False
-        fb_lid_14 = forest_b[leaf_id][14]
-        if regrf:
+        leaf_id = mcf_forest_add.get_terminal_leaf_no(forest_b, x_dat_i)
+        if forest_b[leaf_id][14] is None:  # Leave will be ignored
+            empty_leaf, weights_ij_np = True, 0
+        else:
+            empty_leaf = False
+    if not empty_leaf:
+        if continuous:
+            fb_lid_14_list = [forest_b[leaf_id][14]
+                              for leaf_id in leaf_id_list]
+        else:
+            fb_lid_14 = forest_b[leaf_id][14]
+        if regrf:  # continuous is set to False in previous function
             weights_ij_np = np.zeros((n_y, 1))
-            # n_x_i = len(forest_b[leaf_id][14])
-            # weights_ij_np[forest_b[leaf_id][14], 0] += 1 / n_x_i
             n_x_i = len(fb_lid_14)
             weights_ij_np[fb_lid_14, 0] += 1 / n_x_i
         else:
             weights_ij_np = np.zeros((n_y, no_of_treat))
-            # d_ib = d_dat[forest_b[leaf_id][14]].reshape(-1)  # view
-            d_ib = d_dat[fb_lid_14].reshape(-1)  # view
-            leaf_complete = True
-            for jdx, treat in enumerate(d_values):
-                indices_ibj = d_ib == treat
-                if np.any(indices_ibj):  # any valid observations?
-                    # n_x_i = len(forest_b[leaf_id][14][indices_ibj])
-                    # weights_ij_np[forest_b[leaf_id][14][indices_ibj],
-                    #               jdx] += 1 / n_x_i
-                    fb_lid_14_indi = fb_lid_14[indices_ibj]
-                    n_x_i = len(fb_lid_14_indi)
-                    weights_ij_np[fb_lid_14_indi, jdx] += 1 / n_x_i
-                else:
-                    leaf_complete = False
-                    break
+            if continuous:
+                # We need to collect information over various leafs for the 0
+                # For the other leaves, the leaves are treatment specific
+                leaf_0_complete, leaf_pos_complete = False, True
+                for jdx, _ in enumerate(fb_lid_14_list):  # Zuerst 0 einsammeln
+                    fb_lid_14 = fb_lid_14_list[jdx]
+                    d_ib = d_dat[fb_lid_14].reshape(-1)  # view
+                    indices_ibj_0 = d_ib < 1e-15
+                    indices_ibj_pos = d_ib >= 1e-15
+                    if np.any(indices_ibj_0):  # any valid observations?
+                        fb_lid_14_indi_0 = fb_lid_14[indices_ibj_0]
+                        n_x_i_0 = len(fb_lid_14_indi_0)
+                        weights_ij_np[fb_lid_14_indi_0, 0] += 1 / n_x_i_0
+                        leaf_0_complete = True
+                    if np.any(indices_ibj_pos):  # any valid observations?
+                        fb_lid_14_indi_pos = fb_lid_14[indices_ibj_pos]
+                        n_x_i_pos = len(fb_lid_14_indi_pos)
+                        weights_ij_np[fb_lid_14_indi_pos, jdx+1] += (
+                            1 / n_x_i_pos)
+                    else:
+                        leaf_pos_complete = False
+                        break
+                leaf_complete = leaf_0_complete and leaf_pos_complete
+            else:
+                d_ib = d_dat[fb_lid_14].reshape(-1)  # view
+                leaf_complete = True
+                for jdx, treat in enumerate(d_values):
+                    indices_ibj = d_ib == treat
+                    if np.any(indices_ibj):  # any valid observations?
+                        fb_lid_14_indi = fb_lid_14[indices_ibj]
+                        n_x_i = len(fb_lid_14_indi)
+                        weights_ij_np[fb_lid_14_indi, jdx] += 1 / n_x_i
+                    else:
+                        leaf_complete = False
+                        break
             if not leaf_complete:
-                weights_ij_np = 0
-                empty_leaf = True
+                empty_leaf, weights_ij_np = True, 0
     return weights_ij_np, empty_leaf
 
 
@@ -529,18 +549,14 @@ def final_trans(weights_i_np, no_of_treat, regrf, normalize=True):
 
     """
     if regrf:
-        iterator = 1
-        weights_i = [None]
+        iterator, weights_i = 1, [None]
     else:
-        iterator = no_of_treat
-        weights_i = [None] * no_of_treat
+        iterator, weights_i = no_of_treat, [None] * no_of_treat
     for jdx in range(iterator):
         weights_t = weights_i_np[weights_i_np[:, jdx+1] > 1e-14]
         weights_ti = np.int32(weights_t[:, 0])  # Indices
-        if normalize:
-            weights_tw = weights_t[:, jdx+1] / np.sum(weights_t[:, jdx+1])
-        else:
-            weights_tw = weights_t[:, jdx+1].copy()
+        weights_tw = (weights_t[:, jdx+1] / np.sum(weights_t[:, jdx+1])
+                      if normalize else weights_t[:, jdx+1].copy())
         weights_tw = weights_tw.astype(np.float32)
         weights_i[jdx] = [weights_ti, weights_tw]
         if regrf:
